##
## bubble_sextant.nas
##
##  Nasal code for a bubble sextant.
##
##  Copyright (C) 2007  Anders Gidenstam  (anders(at)gidenstam.org)
##  This file is licensed under the GPL license.
##

var RAD = math.pi/180;
var DEG = 180/math.pi;
var FOV = 25;

## Interface functions.
var toggle_active = func {
    if (getprop("/sim/current-view/internal")) {
        sextant.toggle();
    }
}

# Bind mouseX() and mouseY() to your mode 0 mouse axes in mice.xml. 

var mouseX = func {
    if(__kbd.alt.getValue()) {
        var delta = 3*cmdarg().getNode("offset").getValue();
        if(__kbd.shift.getValue() and sextant.enabled) {
            # Roll adjustment
            var view = "/sim/current-view/roll-offset-deg";
            var val = getprop(view) + delta;
            if(val < -180) val = -180;
            if(val > 180)  val = 180;
            setprop(view, val);
        } else {
            var view = "/sim/current-view/heading-offset-deg";
            var val = getprop(view) - delta;
            if(val < 0)   val = 0;
            if(val > 360) val = 360;
            setprop(view, val);
        }
    }
}

var mouseY = func {
    if(__kbd.alt.getValue()) {
        var delta = 3*cmdarg().getNode("offset").getValue();
        if(__kbd.shift.getValue() and sextant.enabled) {
            # Altitude adjustment   
            var val = sextant.altitude.getValue() - delta;
            if(val < 0) val = 0;
            if(val > 90) val = 90;
            sextant.altitude.setValue(val);
            sextant.reading_deg.setValue(int(val));
            sextant.reading_min.setValue((val - int(val))*60);
        } else {
            var view = "/sim/current-view/pitch-offset-deg";
            var val = getprop(view) - delta;
            if(val < -90) val = -90;
            if(val > 90)  val = 90;
            setprop(view, val);
        }
    }
}


# Singleton class for managing the bubble sextant instrument.
#
var sextant = {

    init : func {
        ## Properties
        me.pitch_err =
          props.globals.getNode("/instrumentation/sextant/pitch-error-deg", 1);
        me.pitch_err.setDoubleValue(0);
        me.roll_err =
          props.globals.getNode("/instrumentation/sextant/roll-error-deg", 1);
        me.roll_err.setDoubleValue(0);
        me.altitude =
          props.globals.getNode("/instrumentation/sextant/altitude-deg", 1);
        me.altitude.setDoubleValue(0);
        me.reading_min =
          props.globals.getNode("/instrumentation/sextant/reading/min",
          1);
        me.reading_min.setDoubleValue(0);
        me.reading_deg =
          props.globals.getNode("/instrumentation/sextant/reading/deg",
          1);
        me.reading_deg.setDoubleValue(0);
        me.servicable =
          props.globals.getNode("/instrumentation/sextant/servicable", 1);
        me.servicable.setBoolValue(1);
        me.enabled = 0;
        me.loopid = 0;

        print("BubbleSextant ... initialized");
    },
    toggle : func {
        if (me.enabled) {
            me.disable();
        } else {
            me.enable();
        }
    },
    enable : func {
        me.enabled = 1;
        setprop("/sim/hud/enable3d[1]", 0);
        setprop("/sim/hud/visibility[1]", 1);

        me.old_view = view.point.save();
        setprop("/sim/current-view/field-of-view", FOV);
        me.old_hud_color = getprop("/sim/hud/current-color");
        setprop("/sim/hud/current-color", 2);

        me.loopid += 1;
        me._loop_(me.loopid);
    },
    disable : func {
        me.enabled = 0;
        setprop("/sim/hud/visibility[1]", 0);
        setprop("/sim/hud/current-color", me.old_hud_color);
        setprop("/sim/current-view/field-of-view",
                me.old_view.getChild("field-of-view").getValue());
        view.point.restore();
    },
    update : func {
        ## State data we need.
        var heading_ac = getprop("/orientation/heading-deg") * RAD;
        var pitch_ac   = getprop("/orientation/pitch-deg") * RAD;
        var roll_ac    = getprop("/orientation/roll-deg") * RAD;
        var yaw_v      =
            getprop("/sim/current-view/heading-offset-deg") * RAD;
        var pitch_v    = getprop("/sim/current-view/pitch-offset-deg") * RAD;
        var roll_v     = getprop("/sim/current-view/roll-offset-deg") * RAD;

        ##  Compute local aircraft axes vectors in the local frame
        ##    Account for aircraft orientation. (x/y/z = front/left/up)
        var T_ac = mulMM(mulMM(rotateZ(heading_ac), rotateY(pitch_ac)),
                         rotateX(roll_ac));
        ##  Account for view orientation and sextant settings.
        ##    The sextant frame is assumed to coincide with the view frame
        ##    except that it is pitched down altitude deg around its local
        ##    Y axis.
        T_bs =  mulMM(mulMM(mulMM(mulMM(
                    T_ac,
                    rotateZ(yaw_v)),
                    rotateY(pitch_v)),
                    rotateX(roll_v)),
                    rotateY(-me.altitude.getValue() * RAD));

        var X_bs = mulMv(T_bs, X);
        var Y_bs = mulMv(T_bs, Y);
        var Z_bs = mulMv(T_bs, Z);

        ## Transform up in the local frame to the sextant frame.
        var Up = mulMv([X_bs, Y_bs, Z_bs], Z);
        var Up_xz = [Up[0], 0, Up[2]];
        var Up_yz = [0, Up[1], Up[2]];

        ## Compute interesting angles in the sextant frame.
        var p_err = angleV(Z, Up_xz);
        if (scalar(Up_xz, X) < 0.0) {
             p_err *= -1;
        }
        me.pitch_err.setValue(p_err*DEG);

        var r_err = angleV(Z, Up_yz);
        if (scalar(Up_yz, Y) < 0.0) {
             r_err *= -1;
        }
        me.roll_err.setValue(r_err*DEG);
    },
    _loop_ : func(id) {
        id == me.loopid and me.enabled or return;
        me.update();
        settimer(func { me._loop_(id); }, 0);
    }

};

setlistener("/sim/signals/fdm-initialized", func {
    sextant.init();
});

###############################################################################
## Ugly matrix math as needed.
## Probably horribly inefficient matrix representation:
##   M[row][col] = [[row1], [row2], [row3]]
##   v[row] = [x, y, z]

var X = [1, 0, 0];
var Y = [0, 1, 0];
var Z = [0, 0, 1];
var id = [X, Y, Z];

var mulMv = func(M, v) {
    return [M[0][0]*v[0] + M[0][1]*v[1] + M[0][2]*v[2],
            M[1][0]*v[0] + M[1][1]*v[1] + M[1][2]*v[2],
            M[2][0]*v[0] + M[2][1]*v[1] + M[2][2]*v[2]];
}

var mulMM = func(A, B) {
    return [[A[0][0]*B[0][0] + A[0][1]*B[1][0] + A[0][2]*B[2][0],
             A[0][0]*B[0][1] + A[0][1]*B[1][1] + A[0][2]*B[2][1],
             A[0][0]*B[0][2] + A[0][1]*B[1][2] + A[0][2]*B[2][2]],
            [A[1][0]*B[0][0] + A[1][1]*B[1][0] + A[1][2]*B[2][0],
             A[1][0]*B[0][1] + A[1][1]*B[1][1] + A[1][2]*B[2][1],
             A[1][0]*B[0][2] + A[1][1]*B[1][2] + A[1][2]*B[2][2]],
            [A[2][0]*B[0][0] + A[2][1]*B[1][0] + A[2][2]*B[2][0],
             A[2][0]*B[0][1] + A[2][1]*B[1][1] + A[2][2]*B[2][1],
             A[2][0]*B[0][2] + A[2][1]*B[1][2] + A[2][2]*B[2][2]]];
}

var scalar = func(a, b) {
    return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}

var absV = func(a) {
    return math.sqrt(scalar(a, a));
}

var crossV = func(a, b) {
    return [a[1]*b[2] - a[2]*b[1],
            a[2]*b[0] - a[0]*b[2],
            a[0]*b[1] - a[1]*b[0]];
}

var rotateX = func (r) {
    return [[1, 0, 0],
            [0, math.cos(-r), -math.sin(-r)],
            [0, math.sin(-r), math.cos(-r)]];
}

var rotateY = func (r) {
    return [[math.cos(-r), 0, math.sin(-r)],
            [0, 1, 0],
            [-math.sin(-r), 0, math.cos(-r)]];
}

var rotateZ = func (r) {
    return [[math.cos(r), -math.sin(r), 0],
            [math.sin(r), math.cos(r), 0],
            [0, 0, 1]];
}

var printMat = func (m) {
    foreach (e; m) {
        if (typeof(e) == "scalar") {
            print(" " ~ e);
        } else {
            var line = ""; 
            foreach (ee; e) {
                line = line ~ " " ~ ee;
            }
            print(line);
        }
    }
}

var angleV = func (a, b) {
    return math.acos(math.abs(scalar(a,b)) / (absV(a) * absV(b)));
}

#print("id*id=");
#printMat(mulMM(id,id));

#print("X=");
#printMat(X);

#print("RotY(45)*X=");
#printMat(mulMv(rotateY(45*math.pi/180),X));

#print("Y*X= " ~scalar(Y,X));
